#if defined(X3J11) || defined(__STDC__)
typedef void    *pointer;               /* generic pointer type */
#else
typedef char    *pointer;               /* generic pointer type */
#endif

#define NULL    0                       /* null pointer constant */
#define xmalloc malloc

extern void     free();
extern pointer  xmalloc();




#ifndef STACK_DIRECTION
#define STACK_DIRECTION 0               /* direction unknown */
#endif

#if STACK_DIRECTION != 0

#define STACK_DIR       STACK_DIRECTION /* known at compile-time */

#else   /* STACK_DIRECTION == 0; need run-time code */

static int      stack_dir;              /* 1 or -1 once known */
#define STACK_DIR       stack_dir

static void
find_stack_direction (/* void */)
{
  static char   *addr = NULL;   /* address of first
                                   `dummy', once known */
  auto char     dummy;          /* to get stack address */

  if (addr == NULL)
    {                           /* initial entry */
      addr = &dummy;

      find_stack_direction ();  /* recurse once */
    }
  else                          /* second entry */
    if (&dummy > addr)
      stack_dir = 1;            /* stack grew upward */
    else
      stack_dir = -1;           /* stack grew downward */
}

#endif  /* STACK_DIRECTION == 0 */

/*
        An "alloca header" is used to:
        (a) chain together all alloca()ed blocks;
        (b) keep track of stack depth.

        It is very important that sizeof(header) agree with malloc()
        alignment chunk size.  The following default should work okay.
*/

#ifndef ALIGN_SIZE
#define ALIGN_SIZE      sizeof(double)
#endif

typedef union hdr
{
  char  align[ALIGN_SIZE];      /* to force sizeof(header) */
  struct
    {
      union hdr *next;          /* for chaining headers */
      char *deep;               /* for stack depth measure */
    } h;
} header;

static header *last_alloca_header = NULL; /* -> last alloca header */

pointer
alloca (size)                   /* returns pointer to storage */
     unsigned   size;           /* # bytes to allocate */
{
  auto char     probe;          /* probes stack depth: */
  register char *depth = &probe;

#if STACK_DIRECTION == 0
  if (STACK_DIR == 0)           /* unknown growth direction */
    find_stack_direction ();
#endif

                                /* Reclaim garbage, defined as all alloca()ed s
torage that
                                   was allocated from deeper in the stack than 
currently. */

  {
    register header     *hp;    /* traverses linked list */

    for (hp = last_alloca_header; hp != NULL;)
      if (STACK_DIR > 0 && hp->h.deep > depth
          || STACK_DIR < 0 && hp->h.deep < depth)
        {
          register header       *np = hp->h.next;

          free ((pointer) hp);  /* collect garbage */

          hp = np;              /* -> next header */
        }
      else
        break;                  /* rest are not deeper */

    last_alloca_header = hp;    /* -> last valid storage */
  }

  if (size == 0)
    return NULL;                /* no allocation required */

  /* Allocate combined header + user data storage. */

  {
    register pointer    new = xmalloc (sizeof (header) + size);
    /* address of header */

    ((header *)new)->h.next = last_alloca_header;
    ((header *)new)->h.deep = depth;

    last_alloca_header = (header *)new;

    /* User storage begins just after header. */

    return (pointer)((char *)new + sizeof(header));
  }
}
